module ILCSolvers

using NamedTrajectories
using QuantumCollocation
using LinearAlgebra
using Manifolds

using ..Experiments
using ..ILCProblems

function QuantumCollocation.solve!(
    prob::ILCProblem;
    io::IO=stdout,
    verbose=true,
    unitary=true,
    manifold_projection=true
)
    i = 1

    while (i == 1 ? true : prob.objective_history[end] > prob.settings[:tol]) &&
        i ≤ prob.settings[:max_iter]

        if verbose
            println(io)
            println(io, "Iteration ", i, "\n")
        end

        Zⁱ = prob.Zs[i]
        Aⁱ = Zⁱ[prob.drive_name]
        dAⁱ = Zⁱ[prob.drive_deriv_name]
        Δtⁱ = timesteps(Zⁱ)
        Yⁱ = prob.experiment(Aⁱ, Δtⁱ)
        Jⁱ = prob.optimizer.J(Yⁱ.ys[end])

        if verbose
            println(io, "\n    J(Aⁱ) = $Jⁱ\n")
            flush(io)
        end

        Yⁱ = prob.continuous_traj ? prob.experiment(Aⁱ, dAⁱ, Δtⁱ) : prob.experiment(Aⁱ, Δtⁱ)

        ΔY = Yⁱ - prob.Y_goal

        dZⁱ = prob.QP(ΔY, Zⁱ, io)

        Zⁱ⁺¹ = prob.optimizer(Zⁱ, dZⁱ, Jⁱ, prob.drive_name, prob.experiment; io=io, manifold_projection=manifold_projection, unitary=unitary)

        push!(prob.Ys, Yⁱ)
        push!(prob.objective_history, Jⁱ)
        push!(prob.dZs, dZⁱ)
        push!(prob.Zs, Zⁱ⁺¹)

        i += 1
    end
    flush(io)
end




end
